import warnings
from decimal import Decimal
from itertools import chain

from django.conf import settings
from django.contrib import messages
from django.core.exceptions import ImproperlyConfigured
from django.db.models import Q
from django.http import Http404, HttpResponseForbidden, HttpResponseRedirect
from django.shortcuts import get_object_or_404, render
from django.urls import reverse, reverse_lazy
from django.utils.translation import gettext_lazy as _
from django.views.generic import CreateView, RedirectView, TemplateView, View
from django.views.generic.detail import (
    DetailView,
    SingleObjectMixin,
    SingleObjectTemplateResponseMixin,
)
from django.views.generic.edit import (
    DeleteView,
    FormView,
    ModelFormMixin,
    ProcessFormView,
)
from django.views.generic.list import ListView
from next_url_mixin.mixin import NextUrlMixin

from plans.base.models import (
    AbstractBillingInfo,
    AbstractInvoice,
    AbstractOrder,
    AbstractPlan,
    AbstractPlanPricing,
    AbstractQuota,
    AbstractUserPlan,
)
from plans.forms import BillingInfoForm, CreateOrderForm, FakePaymentsForm
from plans.importer import import_name
from plans.mixins import LoginRequired
from plans.plan_change import get_change_price
from plans.signals import order_started
from plans.utils import get_currency
from plans.validators import plan_validation

UserPlan = AbstractUserPlan.get_concrete_model()
PlanPricing = AbstractPlanPricing.get_concrete_model()
Plan = AbstractPlan.get_concrete_model()
Order = AbstractOrder.get_concrete_model()
BillingInfo = AbstractBillingInfo.get_concrete_model()
Quota = AbstractQuota.get_concrete_model()
Invoice = AbstractInvoice.get_concrete_model()


class AccountActivationView(LoginRequired, TemplateView):
    template_name = "plans/account_activation.html"

    def get_context_data(self, **kwargs):
        if (
            self.request.user.userplan.active is True
            or self.request.user.userplan.is_expired()
        ):
            raise Http404()

        context = super(AccountActivationView, self).get_context_data(**kwargs)
        errors = self.request.user.userplan.clean_activation()

        if errors["required_to_activate"]:
            context["SUCCESSFUL"] = False
        else:
            context["SUCCESSFUL"] = True
            messages.success(self.request, _("Your account is now active"))

        for error in errors["required_to_activate"]:
            messages.error(self.request, error)
        for error in errors["other"]:
            messages.warning(self.request, error)

        return context


class PlanTableMixin(object):
    def get_plan_table(self, plan_list):
        """
        This method return a list in following order:
        [
            ( Quota1, [ Plan1Quota1, Plan2Quota1, ... , PlanNQuota1] ),
            ( Quota2, [ Plan1Quota2, Plan2Quota2, ... , PlanNQuota2] ),
            ...
            ( QuotaM, [ Plan1QuotaM, Plan2QuotaM, ... , PlanNQuotaM] ),
        ]

        This can be very easily printed as an HTML table element with quotas by row.

        Quotas are calculated based on ``plan_list``. These are all available quotas that are
        used by given plans. If any ``Plan`` does not have any of ``PlanQuota`` then value ``None``
        will be propagated to the data structure.

        """

        # Retrieve all quotas that are used by any ``Plan`` in ``plan_list``
        quota_list = (
            Quota.objects.all().filter(planquota__plan__in=plan_list).distinct()
        )

        # Create random access dict that for every ``Plan`` map ``Quota`` -> ``PlanQuota``
        plan_quotas_dic = {}
        for plan in plan_list:
            plan_quotas_dic[plan] = {}
            for plan_quota in plan.planquota_set.all():
                plan_quotas_dic[plan][plan_quota.quota] = plan_quota

        # Generate data structure described in method docstring, propagate ``None`` whenever
        # ``PlanQuota`` is not available for given ``Plan`` and ``Quota``
        return map(
            lambda quota: (
                quota,
                map(lambda plan: plan_quotas_dic[plan].get(quota, None), plan_list),
            ),
            quota_list,
        )


class PlanTableViewBase(PlanTableMixin, ListView):
    model = Plan
    context_object_name = "plan_list"

    def get_queryset(self):
        queryset = (
            super(PlanTableViewBase, self)
            .get_queryset()
            .prefetch_related("planpricing_set__pricing", "planquota_set__quota")
        )
        if self.request.user.is_authenticated:
            queryset = queryset.filter(
                Q(available=True, visible=True)
                & (Q(customized=self.request.user) | Q(customized__isnull=True))
            )
        else:
            queryset = queryset.filter(
                Q(available=True, visible=True) & Q(customized__isnull=True)
            )
        return queryset

    def get_context_data(self, **kwargs):
        context = super(PlanTableViewBase, self).get_context_data(**kwargs)

        if self.request.user.is_authenticated:
            try:
                self.userplan = UserPlan.objects.select_related("plan").get(
                    user=self.request.user
                )
            except UserPlan.DoesNotExist:
                self.userplan = None

            context["userplan"] = self.userplan

            try:
                context["current_userplan_index"] = list(self.object_list).index(
                    self.userplan.plan
                )
            except (ValueError, AttributeError):
                pass

        context["plan_table"] = self.get_plan_table(self.object_list)
        context["CURRENCY"] = settings.PLANS_CURRENCY

        return context


class CurrentPlanView(LoginRequired, PlanTableViewBase):
    template_name = "plans/current.html"

    def get_queryset(self):
        return Plan.objects.filter(userplan__user=self.request.user).prefetch_related(
            "planpricing_set__pricing", "planquota_set__quota"
        )


class UpgradePlanView(LoginRequired, PlanTableViewBase):
    template_name = "plans/upgrade.html"


class PricingView(PlanTableViewBase):
    template_name = "plans/pricing.html"


class ChangePlanView(LoginRequired, View):
    """
    A view for instant changing user plan when it does not require additional payment.
    Plan can be changed without payment when:
    * user can enable this plan (it is available & visible and if it is customized it is for him,
    * plan is different from the current one that user have,
    * within current change plan policy this does not require any additional payment (None)

    It always redirects to ``upgrade_plan`` url as this is a potential only one place from
    where change plan could be invoked.
    """

    def get(self, request, *args, **kwargs):
        return HttpResponseRedirect(reverse("upgrade_plan"))

    def post(self, request, *args, **kwargs):
        plan = get_object_or_404(
            Plan,
            Q(pk=kwargs["pk"])
            & Q(available=True, visible=True)
            & (Q(customized=request.user) | Q(customized__isnull=True)),
        )
        if request.user.userplan.plan != plan:
            policy = import_name(
                getattr(
                    settings,
                    "PLANS_CHANGE_POLICY",
                    "plans.plan_change.StandardPlanChangePolicy",
                )
            )()

            period = request.user.userplan.days_left()
            price = policy.get_change_price(request.user.userplan.plan, plan, period)

            if price is None:
                request.user.userplan.extend_account(plan, None)
                messages.success(request, _("Your plan has been successfully changed"))
            else:
                return HttpResponseForbidden()
        return HttpResponseRedirect(reverse("upgrade_plan"))


class CreateOrderView(LoginRequired, CreateView):
    template_name = "plans/create_order.html"
    form_class = CreateOrderForm

    def recalculate(self, amount, billing_info):
        """
        Calculates and return pre-filled Order
        """
        order = Order(pk=-1)
        order.recalculate(amount, billing_info, self.request)
        return order

    def validate_plan(self, plan):
        validation_errors = plan_validation(self.request.user, plan)
        if validation_errors["required_to_activate"] or validation_errors["other"]:
            messages.error(
                self.request,
                _(
                    "The selected plan is insufficient for your account. "
                    "Your account will not be activated or will not work fully after completing this order."
                    "<br><br>Following limits will be exceeded: <ul><li>%(reasons)s</ul>"
                )
                % {
                    "reasons": "<li>".join(
                        chain(
                            validation_errors["required_to_activate"],
                            validation_errors["other"],
                        )
                    ),
                },
            )

    def get_all_context(self):
        """
        Retrieves Plan and Pricing for current order creation
        """
        self.plan_pricing = get_object_or_404(
            PlanPricing.objects.all().select_related("plan", "pricing"),
            Q(pk=self.kwargs["pk"])
            & Q(plan__available=True)
            & (
                Q(plan__customized=self.request.user) | Q(plan__customized__isnull=True)
            ),
        )

        # User is not allowed to create new order for Plan when he has different Plan
        # unless it's a free plan. Otherwise, the should use Plan Change View for this
        # kind of action
        if (
            not self.request.user.userplan.is_expired()
            and not self.request.user.userplan.plan.is_free()
            and self.request.user.userplan.plan != self.plan_pricing.plan
        ):
            raise Http404

        self.plan = self.plan_pricing.plan
        self.pricing = self.plan_pricing.pricing

    def get_billing_info(self):
        try:
            return self.request.user.billinginfo
        except BillingInfo.DoesNotExist:
            return None

    def get_price(self):
        return self.plan_pricing.price

    def get_context_data(self, **kwargs):
        context = super(CreateOrderView, self).get_context_data(**kwargs)
        self.get_all_context()
        context["billing_info"] = self.get_billing_info()

        order = self.recalculate(
            self.get_price() or Decimal("0.0"), context["billing_info"]
        )
        order.plan = self.plan_pricing.plan
        order.pricing = self.plan_pricing.pricing
        order.currency = get_currency()
        order.user = self.request.user
        context["object"] = order

        self.validate_plan(order.plan)
        return context

    def form_valid(self, form):
        self.get_all_context()
        order = self.recalculate(
            self.get_price() or Decimal("0.0"), self.get_billing_info()
        )

        self.object = form.save(commit=False)
        self.object.user = self.request.user
        self.object.plan = self.plan
        self.object.pricing = self.pricing
        self.object.amount = order.amount
        self.object.tax = order.tax
        self.object.currency = order.currency
        self.object.save()
        order_started.send(sender=self.object)
        return super(ModelFormMixin, self).form_valid(form)


class CreateOrderPlanChangeView(CreateOrderView):
    template_name = "plans/create_order.html"
    form_class = CreateOrderForm

    def get_all_context(self):
        self.plan = get_object_or_404(
            Plan,
            Q(pk=self.kwargs["pk"])
            & Q(available=True, visible=True)
            & (Q(customized=self.request.user) | Q(customized__isnull=True)),
        )
        self.pricing = None

    def get_price(self):
        return get_change_price(self.request.user.userplan, self.plan)

    def get_context_data(self, **kwargs):
        context = super(CreateOrderView, self).get_context_data(**kwargs)
        self.get_all_context()

        price = self.get_price()
        context["plan"] = self.plan
        context["billing_info"] = self.get_billing_info()
        if price is None:
            context["FREE_ORDER"] = True
            price = 0
        order = self.recalculate(price, context["billing_info"])
        order.pricing = None
        order.plan = self.plan
        order.user = self.request.user
        context["billing_info"] = context["billing_info"]
        context["object"] = order
        self.validate_plan(order.plan)
        return context


class OrderView(LoginRequired, DetailView):
    model = Order

    def get_queryset(self):
        return (
            super(OrderView, self)
            .get_queryset()
            .filter(user=self.request.user)
            .select_related(
                "plan",
                "pricing",
            )
        )


class OrderListView(LoginRequired, ListView):
    model = Order
    paginate_by = 10

    def get_context_data(self, **kwargs):
        context = super(OrderListView, self).get_context_data(**kwargs)
        self.CURRENCY = getattr(settings, "PLANS_CURRENCY", None)
        if len(self.CURRENCY) != 3:
            raise ImproperlyConfigured(
                "PLANS_CURRENCY should be configured as 3-letter currency code."
            )
        context["CURRENCY"] = self.CURRENCY
        return context

    def get_queryset(self):
        return (
            super(OrderListView, self)
            .get_queryset()
            .filter(user=self.request.user)
            .select_related(
                "plan",
                "pricing",
            )
        )


class OrderPaymentReturnView(LoginRequired, DetailView):
    """
    This view is a fallback from any payments processor. It allows just to set additional message
    context and redirect to Order view itself.
    """

    model = Order
    status = None

    def render_to_response(self, context, **response_kwargs):
        if self.status == "success":
            messages.success(
                self.request,
                _(
                    "Thank you for placing a payment. It will be processed as soon as possible."
                ),
            )
        elif self.status == "failure":
            messages.error(
                self.request,
                _(
                    "Payment was not completed correctly. Please repeat payment process."
                ),
            )

        return HttpResponseRedirect(self.object.get_absolute_url())

    def get_queryset(self):
        return (
            super(OrderPaymentReturnView, self)
            .get_queryset()
            .filter(user=self.request.user)
        )


class SuccessUrlMixin:
    def get_success_url(self):
        messages.success(self.request, _("Billing info has been updated successfuly."))
        return reverse("billing_info")


class CreateOrUpdateView(
    SingleObjectTemplateResponseMixin, ModelFormMixin, ProcessFormView
):
    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().post(request, *args, **kwargs)


class BillingInfoCreateOrUpdateView(
    NextUrlMixin, SuccessUrlMixin, LoginRequired, CreateOrUpdateView
):
    form_class = BillingInfoForm
    template_name = "plans/billing_info_create_or_update.html"

    def get_object(self):
        try:
            return self.request.user.billinginfo
        except (AttributeError, BillingInfo.DoesNotExist):
            return None

    def form_valid(self, form):
        form.instance.user = self.request.user
        return super().form_valid(form)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs.update(request=self.request)
        return kwargs


class RedirectToBilling(RedirectView):
    url = reverse_lazy("billing_info")
    permanent = False
    query_string = True

    def get_redirect_url(self, *args, **kwargs):
        warnings.warn(
            "This view URL is deprecated. Use plain billing_info instead.",
            DeprecationWarning,
        )
        return super().get_redirect_url(*args, **kwargs)


class BillingInfoDeleteView(LoginRequired, DeleteView):
    """
    Deletes billing data for user
    """

    template_name = "plans/billing_info_delete.html"

    def get_object(self):
        try:
            return self.request.user.billinginfo
        except BillingInfo.DoesNotExist:
            raise Http404

    def get_success_url(self):
        messages.success(self.request, _("Billing info has been deleted."))
        return reverse("billing_info")


class InvoiceDetailView(LoginRequired, DetailView):
    model = Invoice

    def get_template_names(self):
        return getattr(settings, "PLANS_INVOICE_TEMPLATE", "plans/invoices/PL_EN.html")

    def get_context_data(self, **kwargs):
        context = super(InvoiceDetailView, self).get_context_data(**kwargs)
        context["logo_url"] = getattr(settings, "PLANS_INVOICE_LOGO_URL", None)
        context["auto_print"] = True
        return context

    def get_queryset(self):
        if self.request.user.is_superuser:
            return super(InvoiceDetailView, self).get_queryset().select_related("order")
        else:
            return (
                super(InvoiceDetailView, self)
                .get_queryset()
                .filter(user=self.request.user)
                .select_related("order")
            )

    def get(self, request, *args, **kwargs):
        """
        Handle custom invoice 404 page

        Try to explain to multi-account users they need to login to correct
        account to access their invoices.
        """
        try:
            self.object = self.get_object()
        except Http404:
            return render(request, "plans/invoice_404.html", status=404)

        context = self.get_context_data(object=self.object)
        return self.render_to_response(context)


class FakePaymentsView(LoginRequired, SingleObjectMixin, FormView):
    form_class = FakePaymentsForm
    model = Order
    template_name = "plans/fake_payments.html"

    def get_success_url(self):
        return self.object.get_absolute_url()

    def get_queryset(self):
        return (
            super(FakePaymentsView, self).get_queryset().filter(user=self.request.user)
        )

    def dispatch(self, *args, **kwargs):
        if not getattr(settings, "DEBUG", False):
            return HttpResponseForbidden("This view is accessible only in debug mode.")
        self.object = self.get_object()
        return super(FakePaymentsView, self).dispatch(*args, **kwargs)

    def form_valid(self, form):
        if int(form["status"].value()) == Order.STATUS.COMPLETED:
            self.object.complete_order()
            return HttpResponseRedirect(
                reverse("order_payment_success", kwargs={"pk": self.object.pk})
            )
        else:
            self.object.status = form["status"].value()
            self.object.save()
            return HttpResponseRedirect(
                reverse("order_payment_failure", kwargs={"pk": self.object.pk})
            )



# Generated by Django 3.0.13 on 2021-03-03 10:34

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("plans", "0008_recurringuserplan_token_verified"),
    ]

    operations = [
        migrations.AlterField(
            model_name="planquota",
            name="value",
            field=models.BigIntegerField(blank=True, default=1, null=True),
        ),
    ]



# Generated by Django 2.0.6 on 2018-08-24 14:21

from django.conf import settings
from django.db import migrations


def create_userplans(apps, schema_editor):
    """
    Create UserPlan instances for users that don't have one.

    This function uses the historical model state to avoid issues with
    fields that don't exist yet in the database schema (like updated_at
    which is added in migration 0011).
    """
    User = apps.get_model(settings.AUTH_USER_MODEL)
    Plan = apps.get_model("plans", "Plan")
    UserPlan = apps.get_model("plans", "UserPlan")

    # Get the default plan
    try:
        default_plan = Plan.objects.filter(default=True).first()
    except Plan.DoesNotExist:
        default_plan = None

    if default_plan is None:
        # No default plan exists, skip creating user plans
        return

    # Find users without user plans
    users_without_plans = User.objects.filter(userplan=None)

    # Create user plans for these users
    for user in users_without_plans:
        UserPlan.objects.create(
            user=user,
            plan=default_plan,
            active=False,
            expire=None,
        )


class Migration(migrations.Migration):
    dependencies = [
        ("plans", "0003_make_plans_unique"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.RunPython(create_userplans, reverse_code=migrations.RunPython.noop),
    ]



from django.contrib.auth import get_user_model
from django.test import TestCase, override_settings
from django.urls import reverse
from model_bakery import baker

User = get_user_model()


def print_response(response):
    with open("response.html", "wb") as f:
        f.write(response.content)


class AccountActivationViewTests(TestCase):
    def test_get_active_plan(self):
        user = baker.make(User)
        baker.make("UserPlan", user=user)
        self.client.force_login(user)
        response = self.client.get(reverse("account_activation"))
        self.assertEqual(response.status_code, 404)

    def test_get(self):
        user = baker.make(User)
        baker.make("UserPlan", user=user, active=False)
        self.client.force_login(user)
        response = self.client.get(reverse("account_activation"))
        self.assertContains(response, "<h1>Activation successful</h1>", html=True)


class CurrentPlanViewTests(TestCase):
    def test_get(self):
        user = baker.make(User, username="Foo")
        baker.make("UserPlan", user=user, active=False)
        self.client.force_login(user)
        response = self.client.get(reverse("current_plan"))
        self.assertContains(response, "<h1>Your account</h1>", html=True)
        self.assertContains(response, "<dd>Foo</dd>", html=True)


class ChangePlanViewTests(TestCase):
    def test_get(self):
        user = baker.make(User, username="Foo")
        user_plan = baker.make("UserPlan", user=user, active=False)
        self.client.force_login(user)
        response = self.client.get(
            reverse("change_plan", kwargs={"pk": user_plan.plan.pk})
        )
        self.assertRedirects(response, "/plan/upgrade/")

    def test_post(self):
        user = baker.make(User, username="Foo")
        user_plan = baker.make("UserPlan", user=user, active=False)
        plan = baker.make("Plan", available=True, visible=True)
        self.client.force_login(user)
        response = self.client.post(reverse("change_plan", kwargs={"pk": plan.pk}))
        self.assertRedirects(response, "/plan/upgrade/")
        user_plan.refresh_from_db()
        self.assertEqual(user_plan.plan, plan)


class CreateOrderPlanChangeViewTests(TestCase):
    def test_get(self):
        user = baker.make(User, username="Foo")
        baker.make("UserPlan", user=user, active=False)
        plan = baker.make("Plan", available=True, visible=True, name="Foo plan")
        self.client.force_login(user)
        response = self.client.get(
            reverse("create_order_plan_change", kwargs={"pk": plan.pk})
        )
        self.assertContains(response, "<h1>Confirm order</h1>", html=True)
        self.assertContains(response, "<td>Plan Foo plan (upgrade) </td>", html=True)


class OrderViewTests(TestCase):
    def test_get(self):
        user = baker.make(User, username="Foo")
        order = baker.make("Order", user=user, plan__name="Foo plan")
        self.client.force_login(user)
        response = self.client.get(reverse("order", kwargs={"pk": order.pk}))
        self.assertContains(
            response, f"<h1>Order #{order.pk} (status: new)</h1>", html=True
        )
        self.assertContains(response, "<td>Plan Foo plan (upgrade) </td>", html=True)


class OrderListViewTests(TestCase):
    def test_get(self):
        user = baker.make(User, username="Foo")
        order = baker.make("Order", user=user, plan__name="Foo plan")
        self.client.force_login(user)
        response = self.client.get(reverse("order_list"))
        self.assertContains(response, "<h1>List of orders</h1>", html=True)
        self.assertContains(
            response, f'<a href="/plan/order/{order.pk}/">{order.pk}</a>', html=True
        )
        self.assertContains(
            response,
            f'<a href="/plan/order/{order.pk}/">Plan Foo plan (upgrade) </a>',
            html=True,
        )


class OrderPaymentReturnViewTests(TestCase):
    def test_success(self):
        user = baker.make(User, username="Foo")
        order = baker.make("Order", user=user, plan__name="Foo plan")
        self.client.force_login(user)
        response = self.client.get(
            reverse("order_payment_success", kwargs={"pk": order.pk})
        )
        self.assertRedirects(response, f"/plan/order/{order.pk}/")

    def test_failure(self):
        user = baker.make(User, username="Foo")
        order = baker.make("Order", user=user, plan__name="Foo plan")
        self.client.force_login(user)
        response = self.client.get(
            reverse("order_payment_failure", kwargs={"pk": order.pk})
        )
        self.assertRedirects(response, f"/plan/order/{order.pk}/")


class InvoiceDetailViewTests(TestCase):
    def test_get(self):
        user = baker.make(User, username="Foo")
        order = baker.make("Order", user=user, plan__name="Foo plan")
        invoice = baker.make("Invoice", order=order, user=user, total=123)
        self.client.force_login(user)
        response = self.client.get(
            reverse("invoice_preview_html", kwargs={"pk": invoice.pk})
        )
        self.assertContains(response, '<span class="en">Invoice ID</span>', html=True)
        self.assertContains(
            response, '<td class="number">123.00&nbsp;EUR</td>', html=True
        )

    def test_get_other_user(self):
        user_a = baker.make(User, username="Andy Aight")
        user_b = baker.make(User, username="Bob Bar")
        order_b = baker.make("Order", user=user_b, plan__name="Foo plan")
        invoice_b = baker.make("Invoice", order=order_b, user=user_b, total=123)
        # superusers can access all invoices
        self.assertFalse(user_a.is_superuser)
        self.assertFalse(user_b.is_superuser)
        # login as user_a and try to access user_b's invoice
        self.client.force_login(user_a)
        response = self.client.get(
            reverse("invoice_preview_html", kwargs={"pk": invoice_b.pk})
        )
        self.assertContains(response, "<h1>Invoice Not Found</h1>", status_code=404)
        # login as user B and ensure the invoice is available
        self.client.force_login(user_b)
        response = self.client.get(
            reverse("invoice_preview_html", kwargs={"pk": invoice_b.pk})
        )
        self.assertContains(response, '<span class="en">Invoice ID</span>', html=True)


class FakePaymentViewTests(TestCase):
    def test_get(self):
        user = baker.make(User, username="Foo")
        order = baker.make("Order", user=user, plan__name="Foo plan")
        self.client.force_login(user)
        response = self.client.get(reverse("fake_payments", kwargs={"pk": order.pk}))
        self.assertContains(
            response, "This view is accessible only in debug mode.", status_code=403
        )

    @override_settings(DEBUG=True)
    def test_get_debug(self):
        user = baker.make(User, username="Foo")
        order = baker.make("Order", user=user, plan__name="Foo plan")
        self.client.force_login(user)
        response = self.client.get(reverse("fake_payments", kwargs={"pk": order.pk}))
        self.assertContains(response, "<h1>FakePayments™</h1>", html=True)

    @override_settings(DEBUG=True)
    def test_post_completed(self):
        user = baker.make(User, username="Foo")
        baker.make("UserPlan", user=user, active=False)
        order = baker.make("Order", user=user, plan__name="Foo plan")
        self.client.force_login(user)
        response = self.client.post(
            reverse("fake_payments", kwargs={"pk": order.pk}), {"status": 2}
        )
        self.assertRedirects(
            response,
            f"/plan/order/{order.pk}/payment/success/",
            fetch_redirect_response=False,
        )

    @override_settings(DEBUG=True)
    def test_post_failed(self):
        user = baker.make(User, username="Foo")
        baker.make("UserPlan", user=user, active=False)
        order = baker.make("Order", user=user, plan__name="Foo plan")
        self.client.force_login(user)
        response = self.client.post(
            reverse("fake_payments", kwargs={"pk": order.pk}), {"status": 4}
        )
        self.assertRedirects(
            response,
            f"/plan/order/{order.pk}/payment/failure/",
            fetch_redirect_response=False,
        )



from decimal import Decimal
from unittest.mock import Mock, patch

from django.core.cache import cache
from django.test import TestCase
from requests.exceptions import ConnectionError
from zeep.exceptions import XMLSyntaxError

from plans.taxation.tedb_client import TEDBClient


class TEDBClientTest(TestCase):
    def setUp(self):
        self.client = TEDBClient()
        cache.clear()

    def tearDown(self):
        cache.clear()

    @patch("plans.taxation.tedb_client.Client")
    def test_client_initialization_success(self, mock_client_class):
        """Test successful SOAP client initialization."""
        mock_client = Mock()
        mock_client_class.return_value = mock_client

        client = TEDBClient()

        mock_client_class.assert_called_once_with(TEDBClient.WSDL_URL)
        self.assertEqual(client.client, mock_client)

    @patch("plans.taxation.tedb_client.Client")
    def test_client_initialization_failure(self, mock_client_class):
        """Test SOAP client initialization failure."""
        mock_client_class.side_effect = ConnectionError("Network error")

        client = TEDBClient()

        self.assertIsNone(client.client)

    @patch("plans.taxation.tedb_client.Client")
    def test_client_initialization_xml_syntax_error(self, mock_client_class):
        """Test SOAP client initialization when server returns HTML instead of WSDL."""
        # Simulate the EU server returning an HTML error page
        mock_client_class.side_effect = XMLSyntaxError(
            "Invalid XML content received (xmlParseEntityRef: no name, line 12, column 48)"
        )

        # This should not crash, but gracefully set client to None
        client = TEDBClient()

        self.assertIsNone(client.client)

    def test_cache_key_generation(self):
        """Test cache key generation."""
        key = self.client._get_cache_key("DE", "2025-01-01")
        expected = "tedb_vat_rate_DE_2025-01-01"
        self.assertEqual(key, expected)

    def test_cache_key_generation_default_date(self):
        """Test cache key generation with default date."""
        key = self.client._get_cache_key("DE")
        self.assertTrue(key.startswith("tedb_vat_rate_DE_"))

    @patch("plans.taxation.tedb_client.cache")
    def test_get_vat_rate_from_cache(self, mock_cache):
        """Test retrieving VAT rate from cache."""
        mock_cache.get.return_value = Decimal("19")

        rate = self.client.get_vat_rate("DE")

        self.assertEqual(rate, Decimal("19"))
        mock_cache.get.assert_called_once()

    def test_get_vat_rate_soap_success(self):
        """Test successful VAT rate retrieval from SOAP service."""
        # Mock the client to simulate no SOAP client available
        self.client.client = None

        rate = self.client.get_vat_rate("DE")

        # Should return None (will use static table fallback)
        self.assertIsNone(rate)

    def test_get_vat_rate_soap_network_error(self):
        """Test VAT rate retrieval with network error."""
        # Mock the client to simulate no SOAP client available
        self.client.client = None

        rate = self.client.get_vat_rate("DE")

        self.assertIsNone(rate)

    def test_get_vat_rate_soap_fault_error(self):
        """Test VAT rate retrieval with SOAP fault."""
        # Mock the client to simulate no SOAP client available
        self.client.client = None

        rate = self.client.get_vat_rate("DE")

        self.assertIsNone(rate)

    def test_get_vat_rate_no_client(self):
        """Test VAT rate retrieval when SOAP client is not available."""
        self.client.client = None

        rate = self.client.get_vat_rate("DE")

        self.assertIsNone(rate)

    @patch("plans.taxation.tedb_client.datetime")
    def test_regional_rate_filtering_spain(self, mock_datetime):
        """
        Test that regional rates are filtered correctly (Spain case).

        Spain has 2 STANDARD+DEFAULT rates in TEDB:
        - 7% for Canary Islands (with comment)
        - 21% for mainland (no comment)

        This test verifies we return 21% (mainland), not 7% (regional).
        """
        # Mock current date
        mock_datetime.now.return_value.strftime.return_value = "2025-11-01"

        # Create mock SOAP response structure for Spain
        mock_rate_canary = Mock()
        mock_rate_canary.memberState = "ES"
        mock_rate_canary.type = "STANDARD"
        mock_rate_canary.rate = Mock()
        mock_rate_canary.rate.type = "DEFAULT"
        mock_rate_canary.rate.value = 7.0
        mock_rate_canary.comment = "VAT - Canary Islands - "

        mock_rate_mainland = Mock()
        mock_rate_mainland.memberState = "ES"
        mock_rate_mainland.type = "STANDARD"
        mock_rate_mainland.rate = Mock()
        mock_rate_mainland.rate.type = "DEFAULT"
        mock_rate_mainland.rate.value = 21.0
        mock_rate_mainland.comment = None

        mock_response = Mock()
        mock_response.vatRateResults = [mock_rate_canary, mock_rate_mainland]

        # Mock the SOAP client
        mock_soap_client = Mock()
        mock_soap_client.service.retrieveVatRates.return_value = mock_response
        self.client.client = mock_soap_client

        # Execute
        rate = self.client.get_vat_rate("ES")

        # Verify: Should return 21 (mainland), not 7 (Canary Islands)
        self.assertEqual(rate, Decimal("21"))

        # Verify the rate was cached
        cached = cache.get(self.client._get_cache_key("ES", "2025-11-01"))
        self.assertEqual(cached, Decimal("21"))

    @patch("plans.taxation.tedb_client.datetime")
    def test_all_rates_have_comments_use_highest(self, mock_datetime):
        """
        Test fallback when all rates have comments (e.g., France).

        When all STANDARD+DEFAULT rates have comments, we should use the highest.
        """
        # Mock current date
        mock_datetime.now.return_value.strftime.return_value = "2025-11-01"

        # Create mock response with single rate with comment
        mock_rate = Mock()
        mock_rate.memberState = "FR"
        mock_rate.type = "STANDARD"
        mock_rate.rate = Mock()
        mock_rate.rate.type = "DEFAULT"
        mock_rate.rate.value = 20.0
        mock_rate.comment = "Article 278 of the General Tax Code"

        mock_response = Mock()
        mock_response.vatRateResults = [mock_rate]

        # Mock the SOAP client
        mock_soap_client = Mock()
        mock_soap_client.service.retrieveVatRates.return_value = mock_response
        self.client.client = mock_soap_client

        # Execute
        rate = self.client.get_vat_rate("FR")

        # Verify: Should return 20 (the only/highest rate)
        self.assertEqual(rate, Decimal("20"))

    @patch("plans.taxation.tedb_client.datetime")
    def test_decimal_normalization(self, mock_datetime):
        """Test that whole number rates are normalized (21.0 → 21)."""
        # Mock current date
        mock_datetime.now.return_value.strftime.return_value = "2025-11-01"

        # Create mock response with 21.0 (float from SOAP)
        mock_rate = Mock()
        mock_rate.memberState = "ES"
        mock_rate.type = "STANDARD"
        mock_rate.rate = Mock()
        mock_rate.rate.type = "DEFAULT"
        mock_rate.rate.value = 21.0  # Float value
        mock_rate.comment = None

        mock_response = Mock()
        mock_response.vatRateResults = [mock_rate]

        # Mock the SOAP client
        mock_soap_client = Mock()
        mock_soap_client.service.retrieveVatRates.return_value = mock_response
        self.client.client = mock_soap_client

        # Execute
        rate = self.client.get_vat_rate("ES")

        # Verify: Should be Decimal("21"), not Decimal("21.0")
        self.assertEqual(rate, Decimal("21"))
        self.assertEqual(str(rate), "21")  # Ensure string representation is clean



# Generated by Django 5.2.2 on 2025-06-06 05:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "plans",
            "0014_recurringuserplan_has_automatic_renewal_backup_deprecated_to_renewal_triggered_by",
        ),
    ]

    operations = [
        migrations.AddField(
            model_name="recurringuserplan",
            name="last_renewal_attempt",
            field=models.DateTimeField(
                blank=True, null=True, verbose_name="last renewal attempt"
            ),
        ),
    ]



from swapper import swappable_setting

from plans.base.models import (
    AbstractBillingInfo,
    AbstractInvoice,
    AbstractOrder,
    AbstractPlan,
    AbstractPlanPricing,
    AbstractPlanQuota,
    AbstractPricing,
    AbstractQuota,
    AbstractRecurringUserPlan,
    AbstractUserPlan,
)


class Plan(AbstractPlan):
    class Meta(AbstractPlan.Meta):
        abstract = False
        swappable = swappable_setting("plans", "Plan")


class BillingInfo(AbstractBillingInfo):
    class Meta(AbstractBillingInfo.Meta):
        abstract = False
        swappable = swappable_setting("plans", "BillingInfo")


class UserPlan(AbstractUserPlan):
    class Meta(AbstractUserPlan.Meta):
        abstract = False
        swappable = swappable_setting("plans", "UserPlan")


class Pricing(AbstractPricing):
    class Meta(AbstractPricing.Meta):
        abstract = False
        swappable = swappable_setting("plans", "Pricing")


class PlanPricing(AbstractPlanPricing):
    class Meta(AbstractPlanPricing.Meta):
        abstract = False
        swappable = swappable_setting("plans", "PlanPricing")


class Quota(AbstractQuota):
    class Meta(AbstractQuota.Meta):
        abstract = False
        swappable = swappable_setting("plans", "Quota")


class PlanQuota(AbstractPlanQuota):
    class Meta(AbstractPlanQuota.Meta):
        abstract = False
        swappable = swappable_setting("plans", "PlanQuota")


class Order(AbstractOrder):
    class Meta(AbstractOrder.Meta):
        abstract = False
        swappable = swappable_setting("plans", "Order")


class Invoice(AbstractInvoice):
    class Meta(AbstractInvoice.Meta):
        abstract = False
        swappable = swappable_setting("plans", "Invoice")


class RecurringUserPlan(AbstractRecurringUserPlan):
    class Meta(AbstractRecurringUserPlan.Meta):
        abstract = False
        swappable = swappable_setting("plans", "RecurringUserPlan")



from django.urls import reverse

from plans.base.models import AbstractUserPlan

UserPlan = AbstractUserPlan.get_concrete_model()


def account_status(request):
    """
    Set following ``RequestContext`` variables:

     * ``ACCOUNT_EXPIRED = boolean``, account was expired state,
     * ``ACCOUNT_NOT_ACTIVE = boolean``, set when account is not expired, but it is over quotas so it is
                                        not active
     * ``EXPIRE_IN_DAYS = integer``, number of days to account expiration,
     * ``EXTEND_URL = string``, URL to account extend page.
     * ``ACTIVATE_URL = string``, URL to account activation needed if  account is not active

    """

    if hasattr(request, "user") and request.user.is_authenticated:
        try:
            return {
                "ACCOUNT_EXPIRED": request.user.userplan.is_expired(),
                "ACCOUNT_NOT_ACTIVE": (
                    not request.user.userplan.is_active()
                    and not request.user.userplan.is_expired()
                ),
                "EXPIRE_IN_DAYS": request.user.userplan.days_left(),
                "EXTEND_URL": reverse("current_plan"),
                "ACTIVATE_URL": reverse("account_activation"),
            }
        except UserPlan.DoesNotExist:
            pass
    return {}



from decimal import Decimal
from unittest.mock import patch

from django.test import TestCase, override_settings
from zeep.exceptions import XMLSyntaxError

from plans.taxation.eu import EUTaxationPolicy


class EUTaxationTEDBIntegrationTest(TestCase):
    """Simple integration test for TEDB functionality."""

    @override_settings(PLANS_TAX_COUNTRY="DE")
    def test_fallback_to_static_table_works(self):
        """Test that fallback to static table works when TEDB is unavailable."""
        # This should use static table since TEDB is not fully implemented yet
        rate = EUTaxationPolicy.get_default_tax()

        # Should return Germany's static rate
        self.assertEqual(rate, Decimal("19"))

    @override_settings(PLANS_TAX_COUNTRY="DE")
    def test_get_tax_rate_uses_fallback(self):
        """Test that get_tax_rate uses static table fallback."""
        rate, success = EUTaxationPolicy.get_tax_rate(None, "FR")

        # Should return France's static rate
        self.assertEqual(rate, Decimal("20"))
        self.assertTrue(success)

    @override_settings(PLANS_TAX_COUNTRY="DE")
    def test_tedb_client_initialization(self):
        """Test that TEDB client can be initialized without errors."""
        # This should not raise any exceptions
        client = EUTaxationPolicy._get_tedb_client()
        self.assertIsNotNone(client)

    @override_settings(PLANS_TAX_COUNTRY="DE")
    def test_get_vat_rate_from_tedb_fallback(self):
        """Test TEDB fallback mechanism."""
        # Should fall back to static table
        rate = EUTaxationPolicy._get_vat_rate_from_tedb("DE")
        self.assertEqual(rate, Decimal("19"))

    @override_settings(PLANS_TAX_COUNTRY="DE")
    @patch("plans.taxation.tedb_client.Client")
    def test_fallback_when_tedb_returns_html_error_page(self, mock_client_class):
        """
        Test that system falls back to static rates when TEDB returns HTML error page.

        This simulates the production issue where the EU server returns:
        "Server temporarily unavailable" HTML page instead of WSDL XML.
        """
        # Simulate the EU server returning an HTML error page
        mock_client_class.side_effect = XMLSyntaxError(
            "Invalid XML content received (xmlParseEntityRef: no name, line 12, column 48)"
        )

        # Clear the cached TEDB client to force re-initialization
        if hasattr(EUTaxationPolicy, "_tedb_client"):
            delattr(EUTaxationPolicy, "_tedb_client")

        # This should NOT crash, but fall back to static table
        rate = EUTaxationPolicy._get_vat_rate_from_tedb("DE")

        # Should return Germany's static rate
        self.assertEqual(rate, Decimal("19"))

    @override_settings(PLANS_TAX_COUNTRY="DE")
    @patch("plans.taxation.tedb_client.Client")
    def test_get_tax_rate_falls_back_on_xml_syntax_error(self, mock_client_class):
        """
        Test that get_tax_rate (used in actual user requests) falls back gracefully.

        This is the end-to-end test for the production error scenario.
        """
        # Simulate the EU server returning an HTML error page
        mock_client_class.side_effect = XMLSyntaxError(
            "Invalid XML content received (xmlParseEntityRef: no name, line 12, column 48)"
        )

        # Clear the cached TEDB client to force re-initialization
        if hasattr(EUTaxationPolicy, "_tedb_client"):
            delattr(EUTaxationPolicy, "_tedb_client")

        # This should NOT crash the request, but use static rates
        rate, success = EUTaxationPolicy.get_tax_rate(None, "FR")

        # Should return France's static rate
        self.assertEqual(rate, Decimal("20"))
        self.assertTrue(success)

    @override_settings(PLANS_TAX_COUNTRY="DE")
    def test_updated_vat_rates_in_static_table(self):
        """Test that recently updated VAT rates are correct in static table."""
        # Test recently changed rates
        recent_changes = {
            "EE": Decimal("24"),  # Estonia (changed to 24% in July 2025)
            "FI": Decimal("25.5"),  # Finland (changed from 24% in Sep 2024)
            "SK": Decimal("23"),  # Slovakia (changed from 20% in Jan 2025)
            "RO": Decimal("21"),  # Romania (changed from 19% in Aug 2025)
        }

        for country, expected_rate in recent_changes.items():
            actual_rate = EUTaxationPolicy.EU_COUNTRIES_VAT.get(country)
            self.assertEqual(
                actual_rate,
                expected_rate,
                f"VAT rate for {country} should be {expected_rate}%, got {actual_rate}%",
            )

    @override_settings(PLANS_TAX_COUNTRY="DE")
    def test_static_vat_rates_match_tedb(self):
        """
        Validate that static VAT table matches live TEDB data.

        This test queries the actual TEDB API and compares with our static table.
        It will FAIL if rates don't match, indicating either:
        - VAT rates have changed and need updating
        - Bug in TEDB parsing (e.g., returning regional rates)

        This test is designed to catch issues like Spain returning 7% (Canary Islands)
        instead of 21% (mainland).
        """
        tedb_client = EUTaxationPolicy._get_tedb_client()

        if not tedb_client.client:
            self.skipTest("TEDB client not available")

        # Test all countries in our static table
        for country_code, static_rate in EUTaxationPolicy.EU_COUNTRIES_VAT.items():
            tedb_rate = tedb_client.get_vat_rate(country_code)

            self.assertIsNotNone(
                tedb_rate,
                f"{country_code}: TEDB returned None (service error or no data)",
            )

            self.assertEqual(
                tedb_rate,
                static_rate,
                f"{country_code}: TEDB returned {tedb_rate}% but static table has {static_rate}% "
                f"- either VAT changed or there's a bug (e.g., regional rate like Canary Islands)",
            )



from django.core.management import BaseCommand

from plans.base.models import AbstractUserPlan

UserPlan = AbstractUserPlan.get_concrete_model()


class Command(BaseCommand):
    help = "Creates UserPlans for all Users"

    def handle(self, *args, **options):  # pragma: no cover
        userplans = UserPlan.create_for_users_without_plan()
        self.stdout.write("%s user plans was created" % userplans.count())



# Generated by Django 4.2.11 on 2024-04-10 12:50

from enum import IntEnum

from django.db import migrations


def _recurringuserplan_has_automatic_renewal_backup_deprecated_to_renewal_triggered_by(
    apps, schema_editor
):
    RecurringUserPlan = apps.get_model("plans", "RecurringUserPlan")
    recurringuserplans_changed = (
        RecurringUserPlan.objects.select_for_update()
        .exclude(
            _has_automatic_renewal_backup_deprecated=True,
            renewal_triggered_by=_RenewalTriggeredByEnum.TASK,
        )
        .exclude(
            _has_automatic_renewal_backup_deprecated=False,
            renewal_triggered_by=_RenewalTriggeredByEnum.USER,
        )
    )
    for recurringuserplan_changed in recurringuserplans_changed:
        print(
            "RecurringUserPlan's renewal_triggered_by will be overwritten:",
            recurringuserplan_changed.pk,
        )
    RecurringUserPlan.objects.filter(
        _has_automatic_renewal_backup_deprecated=True
    ).update(renewal_triggered_by=_RenewalTriggeredByEnum.TASK)
    RecurringUserPlan.objects.filter(
        _has_automatic_renewal_backup_deprecated=False
    ).update(renewal_triggered_by=_RenewalTriggeredByEnum.USER)


def _recurringuserplan_renewal_triggered_by_to_has_automatic_renewal_backup_deprecated(
    apps, schema_editor
):
    RecurringUserPlan = apps.get_model("plans", "RecurringUserPlan")
    recurringuserplans_changed = (
        RecurringUserPlan.objects.select_for_update()
        .exclude(
            renewal_triggered_by__in={
                _RenewalTriggeredByEnum.TASK,
                _RenewalTriggeredByEnum.OTHER,
            },
            _has_automatic_renewal_backup_deprecated=True,
        )
        .exclude(
            renewal_triggered_by=_RenewalTriggeredByEnum.USER,
            _has_automatic_renewal_backup_deprecated=False,
        )
    )
    for recurringuserplan_changed in recurringuserplans_changed:
        print(
            "RecurringUserPlan's _has_automatic_renewal_backup_deprecated will be overwritten:",
            recurringuserplan_changed.pk,
        )
    RecurringUserPlan.objects.filter(
        renewal_triggered_by__in={
            _RenewalTriggeredByEnum.TASK,
            _RenewalTriggeredByEnum.OTHER,
        }
    ).update(_has_automatic_renewal_backup_deprecated=True)
    RecurringUserPlan.objects.filter(
        renewal_triggered_by=_RenewalTriggeredByEnum.USER
    ).update(_has_automatic_renewal_backup_deprecated=False)
    RecurringUserPlan.objects.update(renewal_triggered_by=_RenewalTriggeredByEnum.USER)


class _RenewalTriggeredByEnum(IntEnum):
    OTHER = 1
    USER = 2
    TASK = 3


class Migration(migrations.Migration):

    dependencies = [
        ("plans", "0013_alter_recurringuserplan_has_automatic_renewal_and_more"),
    ]

    operations = [
        migrations.RunPython(
            _recurringuserplan_has_automatic_renewal_backup_deprecated_to_renewal_triggered_by,
            reverse_code=_recurringuserplan_renewal_triggered_by_to_has_automatic_renewal_backup_deprecated,
        )
    ]



# Generated manually on 2024-12-01 for adding slug field to Plan model

from django.db import migrations, models
from django.utils.text import slugify


def populate_slugs(apps, schema_editor):
    """Populate slug field from existing name values"""
    Plan = apps.get_model("plans", "Plan")
    used_slugs = set()

    for plan in Plan.objects.all():
        if not plan.slug:
            base_slug = slugify(plan.name)
            # Handle empty slugs from whitespace-only names
            if not base_slug:
                base_slug = f"plan-{plan.pk}"

            slug = base_slug
            counter = 1

            # Handle duplicates by appending a number
            while slug in used_slugs:
                slug = f"{base_slug}-{counter}"
                counter += 1

            plan.slug = slug
            used_slugs.add(slug)
            plan.save()


def reverse_populate_slugs(apps, schema_editor):
    """Reverse migration - clear slug field"""
    Plan = apps.get_model("plans", "Plan")
    Plan.objects.update(slug="")


class Migration(migrations.Migration):

    dependencies = [
        ("plans", "0016_invoice_cancellation_reason_invoice_credit_note_for_and_more"),
    ]

    operations = [
        # Add slug field as nullable CharField first
        migrations.AddField(
            model_name="plan",
            name="slug",
            field=models.CharField(max_length=100, null=True, verbose_name="slug"),
        ),
        # Populate slugs from existing names
        migrations.RunPython(populate_slugs, reverse_populate_slugs),
        # Convert to required SlugField with unique constraint
        # WARNING: This operation creates a unique index synchronously and can
        # LOCK THE TABLE on PostgreSQL with large datasets, causing downtime.
        # For production PostgreSQL deployments, consider:
        # 1. Run: python manage.py migrate plans 0017 --fake
        # 2. Manually: CREATE UNIQUE INDEX CONCURRENTLY plans_plan_slug_unique ON plans_plan (slug);
        # 3. Then: ALTER TABLE plans_plan ADD CONSTRAINT plans_plan_slug_unique UNIQUE USING INDEX plans_plan_slug_unique;
        migrations.AlterField(
            model_name="plan",
            name="slug",
            field=models.SlugField(max_length=100, unique=True, verbose_name="slug"),
        ),
    ]



import datetime
from decimal import Decimal

from django.contrib.auth import get_user_model
from django.test import TestCase, override_settings
from django.utils import timezone
from freezegun import freeze_time
from model_bakery import baker

from plans.models import AbstractRecurringUserPlan, RecurringUserPlan
from plans.tasks import autorenew_account

User = get_user_model()


class AutorenewSchedulerTests(TestCase):
    def setUp(self):
        self.user = baker.make(User, username="test_user", email="test@example.com")
        self.plan = baker.make("Plan", name="Test Plan")
        self.pricing = baker.make("Pricing", period=30)
        baker.make("PlanPricing", plan=self.plan, pricing=self.pricing, price=10)
        self.user_plan = baker.make(
            "UserPlan",
            user=self.user,
            plan=self.plan,
        )
        # Baker's direct creation of related one-to-one objects can be tricky,
        # so we create the RecurringUserPlan explicitly.
        baker.make(
            RecurringUserPlan,
            user_plan=self.user_plan,
            renewal_triggered_by=AbstractRecurringUserPlan.RENEWAL_TRIGGERED_BY.TASK,
            token_verified=True,
            pricing=self.pricing,
            amount=Decimal(10),
            currency="USD",
        )

    @override_settings(PLANS_AUTORENEW_SCHEDULE=[datetime.timedelta(days=3)])
    @freeze_time("2023-01-01 12:00:00")
    def test_autorenew_schedule_before_expiry_should_renew(self):
        """Plan expires in 2 days, schedule is 3 days, should renew."""
        self.user_plan.expire = timezone.now().date() + datetime.timedelta(days=2)
        self.user_plan.save()

        renewed = autorenew_account()
        self.assertEqual(len(renewed), 1)
        self.assertEqual(renewed[0], self.user)

    @override_settings(PLANS_AUTORENEW_SCHEDULE=[datetime.timedelta(days=3)])
    @freeze_time("2023-01-01 12:00:00")
    def test_autorenew_schedule_before_expiry_should_not_renew(self):
        """Plan expires in 4 days, schedule is 3 days, should NOT renew."""
        self.user_plan.expire = timezone.now().date() + datetime.timedelta(days=4)
        self.user_plan.save()

        renewed = autorenew_account()
        self.assertEqual(len(renewed), 0)

    @override_settings(PLANS_AUTORENEW_SCHEDULE=[datetime.timedelta(days=-2)])
    @freeze_time("2023-01-10 12:00:00")
    def test_autorenew_schedule_after_expiry_should_renew(self):
        """Plan expired 3 days ago, schedule is -2 days, should renew."""
        self.user_plan.expire = timezone.now().date() - datetime.timedelta(
            days=3
        )  # expired 2023-01-07
        self.user_plan.save()
        self.user_plan.expire_account()

        renewed = autorenew_account()
        self.assertEqual(len(renewed), 1)
        self.assertEqual(renewed[0], self.user)

    @override_settings(PLANS_AUTORENEW_SCHEDULE=[datetime.timedelta(days=-2)])
    @freeze_time("2023-01-10 12:00:00")
    def test_autorenew_schedule_after_expiry_should_not_renew(self):
        """Plan expired 1 day ago, schedule is -2 days, should NOT renew."""
        self.user_plan.expire = timezone.now().date() - datetime.timedelta(
            days=1
        )  # expired 2023-01-09
        self.user_plan.save()
        self.user_plan.expire_account()

        renewed = autorenew_account()
        self.assertEqual(len(renewed), 0)

    @override_settings(PLANS_AUTORENEW_SCHEDULE=[datetime.timedelta(days=3)])
    @freeze_time("2023-01-01 12:00:00")
    def test_autorenew_schedule_should_not_renew_recently_attempted(self):
        """Plan should not renew if a renewal was attempted recently."""
        self.user_plan.expire = timezone.now().date() + datetime.timedelta(days=2)
        self.user_plan.recurring.last_renewal_attempt = timezone.now()
        self.user_plan.save()
        self.user_plan.recurring.save()

        renewed = autorenew_account()
        self.assertEqual(len(renewed), 0)

    @override_settings(PLANS_AUTORENEW_SCHEDULE=[datetime.timedelta(days=3)])
    @freeze_time("2023-01-10 12:00:00")
    def test_autorenew_schedule_should_renew_attempted_long_ago(self):
        """Plan should renew if a renewal was attempted before the renewal window."""
        self.user_plan.expire = timezone.now().date() + datetime.timedelta(
            days=2
        )  # 2023-01-12
        # expire - schedule = 2023-01-09
        self.user_plan.recurring.last_renewal_attempt = timezone.make_aware(
            datetime.datetime(2023, 1, 8, 11, 59, 59)
        )
        self.user_plan.save()
        self.user_plan.recurring.save()

        renewed = autorenew_account()
        self.assertEqual(len(renewed), 1)
        self.assertEqual(renewed[0], self.user)

    @override_settings(
        PLANS_AUTORENEW_SCHEDULE=[
            datetime.timedelta(days=1),
            datetime.timedelta(days=5),
        ]
    )
    @freeze_time("2023-01-01 12:00:00")
    def test_autorenew_multiple_schedules_should_renew(self):
        """Plan should renew if one of multiple schedules is met."""
        self.user_plan.expire = timezone.now().date() + datetime.timedelta(days=4)
        self.user_plan.save()

        renewed = autorenew_account()
        self.assertEqual(len(renewed), 1)
        self.assertEqual(renewed[0], self.user)

    @override_settings(
        PLANS_AUTORENEW_SCHEDULE=[
            datetime.timedelta(days=1),
            datetime.timedelta(days=2),
        ]
    )
    @freeze_time("2023-01-01 12:00:00")
    def test_autorenew_multiple_schedules_should_not_renew(self):
        """Plan should not renew if no multiple schedules are met."""
        self.user_plan.expire = timezone.now().date() + datetime.timedelta(days=4)
        self.user_plan.save()

        renewed = autorenew_account()
        self.assertEqual(len(renewed), 0)

    @override_settings(
        PLANS_AUTORENEW_SCHEDULE=[
            datetime.timedelta(days=1),
            datetime.timedelta(days=-1),
        ]
    )
    @freeze_time("2023-01-10 12:00:00")
    def test_autorenew_mixed_schedules_after_expiry(self):
        """Plan should renew with mixed (positive/negative) schedules after expiry."""
        self.user_plan.expire = timezone.now().date() - datetime.timedelta(days=1)
        self.user_plan.save()
        self.user_plan.expire_account()

        renewed = autorenew_account()
        self.assertEqual(len(renewed), 1)
        self.assertEqual(renewed[0], self.user)

    @override_settings(PLANS_AUTORENEW_SCHEDULE=[datetime.timedelta(days=3)])
    @freeze_time("2023-01-01 12:00:00", auto_tick_seconds=1)
    def test_autorenew_schedule_does_not_renew_twice(self):
        """A plan should not be renewed twice for the same schedule."""
        self.user_plan.expire = timezone.now().date() + datetime.timedelta(days=2)
        self.user_plan.save()

        # First run, should renew
        renewed_first = autorenew_account()
        self.assertEqual(len(renewed_first), 1)
        self.assertEqual(renewed_first[0], self.user)

        # Second run, should NOT renew
        renewed_second = autorenew_account()
        self.assertEqual(len(renewed_second), 0)

    @override_settings(
        PLANS_AUTORENEW_SCHEDULE=[datetime.timedelta(days=-2)],
        PLANS_AUTORENEW_MAX_DAYS_AFTER_EXPIRY=datetime.timedelta(days=30),
    )
    @freeze_time("2023-01-31 00:00:01")
    def test_autorenew_respects_max_days_after_expiry(self):
        """
        Plan expired exactly on the edge of the max_renew_after window.
        It should be renewed. This test would fail without the `.date()`
        conversion in the task.
        """
        self.user_plan.expire = datetime.date(2022, 12, 30)
        self.user_plan.save()
        self.user_plan.expire_account()

        renewed = autorenew_account()
        self.assertEqual(len(renewed), 1)
        self.assertEqual(renewed[0], self.user)

    @override_settings(
        PLANS_AUTORENEW_SCHEDULE=[datetime.timedelta(days=-2)],
        PLANS_AUTORENEW_MAX_DAYS_AFTER_EXPIRY=datetime.timedelta(days=30),
    )
    @freeze_time("2023-01-31 12:00:00")
    def test_autorenew_respects_max_days_after_expiry_should_not_renew(self):
        """
        Plan expired one day before the max_renew_after window.
        It should NOT be renewed.
        """
        # renewal window starts at 2022-12-30
        self.user_plan.expire = datetime.date(2022, 12, 29)
        self.user_plan.save()
        self.user_plan.expire_account()

        renewed = autorenew_account()
        self.assertEqual(len(renewed), 0)

    @override_settings(
        PLANS_AUTORENEW_SCHEDULE=[datetime.timedelta(days=1)],
        PLANS_AUTORENEW_MAX_DAYS_AFTER_EXPIRY=datetime.timedelta(days=5),
    )
    @freeze_time("2025-06-07 12:10:30")
    def test_autorenew_date_comparison_is_correct(self):
        """
        Tests that comparing a DateField (expire) with a datetime object
        works correctly by using .date() to truncate the time part.
        Without .date(), this test should fail.
        """
        self.user_plan.expire = datetime.date(2025, 6, 7)
        self.user_plan.save()
        self.user_plan.expire_account()

        renewed = autorenew_account()
        self.assertEqual(len(renewed), 1)
        self.assertEqual(renewed[0], self.user)
